#!BPY

#                                                                                
#    Copyright (c) 2005/2006 Thomas Fuchs / The Rockin'-B, www.rockin-b.de       
#                                                                                
#    Permission is granted to anyone to use this software for any purpose        
#    and to redistribute it freely, subject to the following restrictions:       
#                                                                                
#    1.  The origin of this software must not be misrepresented. You must not    
#        claim that you wrote the original software. If you use this software    
#        in a product, an acknowledgment in the product documentation would be   
#        appreciated but is not required.                                        
#                                                                                
#    2.  ANY COMMERCIAL USE IS PROHIBITED. This includes that you must not use   
#        this software in the production of a commercial product. Only an explicit
#        permission by the author can allow you to do that.                      
#                                                                                
#    3.  Any modification applied to the software must be marked as such clearly.
#                                                                                
#    4.  This license may not be removed or altered from any distribution.       
#                                                                                
#                                                                                
#    			    NO WARRANTY                                                 
#                                                                                
#    THERE IS NO WARRANTY FOR THE PROGRAM.                                       
#    THE PROGRAM IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,               
#    EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,                 
#    THE IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR PURPOSE.                 
#    THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
#    SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY    
#    SERVICING, REPAIR OR CORRECTION.                                            
#                                                                                
#    IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES,         
#    INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
#    OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED   
#    TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY    
#    YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER  
#    PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE       
#    POSSIBILITY OF SUCH DAMAGES.                                                
#                                                                                

# How to run this script:
#	Note: the script exports all selected objects, or all objects of the scene, if none are selected
# Way A:
#		1. Put SegaSaturn.py into Blender's scripts directory (MacOS example: /Applications/blender-2.37a/Blender.app/Contents/MacOS/.blender/scripts/)
#		2. Having your 3d scene (example: SegaSaturn.blend) loaded into Blender, open the scripts window (SHIFT+F11, or by window type selector)
#		3. Export: in the scripts window, run the script by mouse-clicking: Scripts->Export->SEGA Saturn SGL (.mdl)... 
# Way B:
#		1. Having your 3d scene (example: SegaSaturn.blend) loaded into Blender, open a text editor window (by window type selector)
#		2. Load the script into the text editor: File->Open and choose SegaSaturn.py
#		3. Export: in the text editor window, press ALT+P
#



# Known bugs:
# - check if exported textures are correct
# - on SEGA Saturn, hierarchy transformations don't seem to be 100% like in Blender

# notes:
# vertex paint: select object, vertex paint mode, in Editing Panel (F9) select color in Paint tab and paint vertices.
#	!! only visible, when the material got VCol Paint selected (Shading Panel (F5), Material tab)


# ToDo:
# - multipass doesn't work, the result is the same afterwards (it's especially needed for the textured 7x7 grid)
# - currently, last textures generated can exhaust more space as first textures generated,
#   so multipass texture height/width computation might be able to equalize that
# - support bicubic filtering for subtexture sampling

# - intelligent uv shifting/rotation to bring longest texture side(s) to width
#   this will influence/permutate poylgon vertex indices, gouraud color list and uv coordinates
# - vram memory allocation, smart placement of gouraud and character pattern tables
# - support more SGL specific attributes: UseLight, .....

# Canned, because not needed:
# - force nearest neighbor sampling for subtextures having exactly the same size as the source texture,
#   such that they are equal


# release notes / change log:
# a plus '+' designates a feature, a minus '-' a bug or missing feature


# 06/12/08:
# + added progress bar
# + settings allowed:
#	via pop-up menus, changed settings are saved in the registry and are automatically restored later
# + added bilinear texture sample filter
# + fixed texture sampling to sample the nearest neighbor
# + added custom integer sqrt, if math package not present
# + for smart texture size determination:
#   texture weight is now it's area, not it's surrounding,
#   texture width and height is now determined by the polygon shape,
#   added setting of maximum for width and height (because yabause doesn't display 504px width textures),
#	max width and height are further restricted by the pixelspace the subtexture occupates on the source image size
#    -> beware, yabause doesn't yet support textures of maximum width 504, but real Saturn does,
#   fixed an error in texture sampling (y_add was wrong), fixed another error which could cause pixels being outside image

# 06/12/04:
# + exports to 3 files:
#   <filename>.mdl  <- 3d object hierarchy polygon data
#   <filename>.gr   <- gouraud tables, definitions and variable declarations (static and real time)
#   <filename>.txr  <- character pattern data, TEXTURE table, PICTURE table
# + texture export:
#   recognizes h/v flipped textures and avoids duplicates,
#   exports character pattern tables, TEXTURE table and PICTURE table,
#   demo supports texture display
# - texture export currently only works for image textures that got UV coordinates assigned (manually)
# + gouraud: switched VRAM position of real and static gouraud,
#   added macro definitions for DMA target address for the gouraud tables,
#   fixed setting of flag CL_Gouraud, it must be set for static and real time gouraud
#   (warning: gouraud does not yet work in yabause 0.7.2, but on real Saturn),
#   added macro definitions and variable declarations needed for real time gouraud,
#   fixed vertex normal export,
#   added static and real time gouraud shading to demo
# + default filenames are taken from the .blend filename instead of always being untitled.mdl
# + discovered & fixed: multiple materials used by a single object work now,
#   even if the material is not contained in the NMesh's list, but in the Object's list
# + fixed: object scale's coordinate system wasn't converted to Saturn coordinate system
# + fixed: converted object rotation from radiants to degrees

# 06/12/01:
# + comes with Saturn demo code (slPutObjectX(), slPutAllObjectsX(), no real time gouraud yet supported)
# + supports objects with no material assigned,
#   then it will create sglDummyMaterial (if it does not already exist) and use that one.
#   The sglDummyMaterial remains unlinked, it's not assigned to any object
# + supports faces with only 3 (triangles) vertices instead of 4 (quads),
#   in this case, vertex 4 is equal to vertex 3
# + converts coordinate systems automatically (OpenGL to SGL):
#   Blender   ->   Saturn
#      x      ->     x
#      y      ->    -z
#      z      ->     y
# + supports hierarchy, uses SGL's OBJECT structure, which got links to children and siblings
# + preserves hierarchy when the selection covers only a subgraph of the hierarchy,
#   for all unselected objects that still need to be present in the hierarchy, only their object
#   transformations are exported, not it's 3d data
# + creates a pointer link to the topmost OBJECT, when starting to render OBJECTS recursively, all OBJECTs can be drawn
# + avoid vntbl if no vertex normals (real time gouraud shading) used in an object
# + supports export of gouraud tables
# + supports linked duplicated objects: equal mesh data used by multiple objects is only exported once
# + does not require a full python installation
# - no texture export yet
# - gouraud feature (and a lot more) untested, only static gouraud created by manually assigning vertex colors
#   can be expected to work correctly
# - script only tested with blender 23.7a on MacOS X
# - rushed out




#Submenu: 'Export all the scene' export
#Submenu: 'Export selected obj' exportsel

""" Registration info for Blender menus:
Name: 'SEGA Saturn SGL (.mdl)...'
Blender: 237a
Group: 'Export'
Tip: 'Export to C format for use with Saturn SGL library.'
"""
__author__ = "Rockin'-B"
__url__ = ("The Rockin'-B website, http://www.rockin-b.de")
__version__ = "06/12/08"

__bpydoc__ = """\
This script exports the Blender mesh object hierarchy (or selected portions of that) to compilable SEGA Saturn SGL C code format.
"""


import Blender
from Blender import Types, NMesh, Material, Texture, Registry, Window

try:
	import math
	from math import sqrt
except:
	def sqrt(a):
		if (a < 1):
			return a
		a2 = a#int(a)
		# avoid integer multiplications, because the warning messages are anoying
		r = 0.0
		for i in range(32):
#			r_next = r + float( (1 << (31-i)) )
			r_next = r + 2.0**(31-i)
			if (r_next*r_next <= a2):
				r = r_next
		if (r < 0 or r*r > a):
			print "ERROR in sqrt: a %s, a2 %s, r %s, r2 %s" % (a, a2, r, r*r)
		return r;

#import os

import sys


def printException(critical, msg):
	print "*** Exception ***"
	print "\tMessage: %s" % (msg)
	if (critical == 0):
		print "\t:-) The output file should work, nevertheless!"
	else:
		print "\t:-( The output file may not work!"
		global nCriticalExceptions
		nCriticalExceptions += 1
	print "\ttype: %s" % (sys.exc_info()[0])
	print "\vvalue: %s" % (sys.exc_info()[1])
	print "\ttraceback: %s" % (sys.exc_info()[2])

				
						
class ExportSGL:

	def __init__(self):
		# Keep track of the things that have already been written.
		# The name in the list might not be the original blender name,
		# but the suffix used in the output
		self.DataNameList = []
#		self.ObjectNameList = []
		# A hash storing a list of children objects for each object.
		# Key is the original blender object name
		self.ChildrenObjectDict = {}
		self.RootObjectList = []
		# which objects to export
		self.SelectedObjectList = []
		# gouraud color tables, each entry is a list of 4 entries of type NMesh.Col()
		self.GrRealTbl = []
		self.GrStaticTbl = []
		# choose some default filenames
		self.setFilenames(Blender.Get('filename'))
		# For each texture to export, this dictionary contains a list of subtextures.
		# Key is the original blender image (for now don't use texture name, because the same image can be used by multiple textures) name.
		# Value is a list [image, SubTexLst] with SubTexLst being a 3 entry list of texture index, uv coordinate lists and weight
		self.TextureDict = {}
		# This uv threshold determines if textures are assumed equal or not
		self.uvThreshold = 1.0 / 512	# works for a maximum source texture resolution (w or h) of 512 pixels
		# the texture index of the next texture
		self.texIndex = 0
		self.texWeightSum = 0
		self.bilinear = 0
		self.maxWidth = 128
		self.maxHeight = 128
		self.maxTexels = (128+64)*1024/2

	def setFilenames(self, filename):
		# filename for gouraud tables file
		fnl = filename.split('.', 1)
#		print "separated list: %s" % (fnl)
		fnl[len(fnl)-1] = "gr"
#		print "modified list: %s" % (fnl)
		self.filename_gr = '.'.join(fnl)
#		print "joined list: %s" % (self.filename_gr)
		fnl[len(fnl)-1] = "mdl"
#		print "modified list: %s" % (fnl)
		self.filename_mdl = '.'.join(fnl)
		fnl[len(fnl)-1] = "txr"
#		print "modified list: %s" % (fnl)
		self.filename_txr = '.'.join(fnl)
		print "filename_gr: %s" % (self.filename_gr)
		print "filename_mdl: %s" % (self.filename_mdl)
		print "filename_txr: %s" % (self.filename_txr)


		
	# As the Blender API only provides Blender.Object.getParent()
	# this enables a depth-first-traversal of the hierarchy by
	# collecting an objects children in a list, which is stored in
	# the dictionary/hash ChildrenObjectDict. The key to access is Object.getName().
	#
	# The depth-first-traversal is supposed to start at root objects, so it
	# collects all root objects in the list RootObjectList.
	
	def findRootAndChildrenObjects(self, nObj):
		print "findRootAndChildrenObjects: start"
		for parent in nObj:
			# register root objects
			if (parent.getParent() == None):
				self.RootObjectList.append(parent)
				print "RootObject %s added" % (parent.getName())
			# search it's child objects
			ChildrenObjectList = []
			for child in nObj:
				if (child.getParent() == parent):
					print "Child %s added to parent %s" % (child.getName(), parent.getName())
					ChildrenObjectList.append(child)
			# add list of children
			self.ChildrenObjectDict[parent.getName()] = ChildrenObjectList
		print "findRootAndChildrenObjects: end\n"
	
	# Performs a depth-first traversal of the object hierarchy.
	
	def exportHierarchy(self, siblings):
		# name_sibling ensures, that an object is only exported,
		# when it is of type NMesh, or one of it's successors (children and children's children)
		# is of type NMesh (then name_sibling is not "")
		name_sibling = ""
		if (len(siblings) != 0):
			
			for s in siblings:
				# export children first
				print "Exporting children of %s (type %s)" % (s.getName(), s.getType())
				name_child = self.exportHierarchy(self.ChildrenObjectDict[s.getName()])
				if(name_child != "" or (self.SelectedObjectList.count(s) == 1 and s.getType() == "Mesh")):
					print "Exporting %s" % s.getName()
					self.writeObject(s, name_child, name_sibling)
					name_sibling = s.getName()
		
		return name_sibling
		
	def exportObjects(self, nObj):
		self.file = open(self.filename_mdl, "w")
		try:
			self.SelectedObjectList = nObj
			self.writeHeader()
			#self.writeNMesh_grreal(nObj, "all")
			self.file.write("// Exporting %s objects\n\n" % len(nObj))

			self.findRootAndChildrenObjects(Blender.Object.Get())#(nObj)
			name_last_object = self.exportHierarchy(self.RootObjectList)
			# add a pointer to easily start traversing the WHOLE object hierarchy that's been written to the file
			self.file.write("OBJECT\t*first_root_object = object_%s;\n" % (name_last_object.replace('.', '_')))
		except:
			self.file.write("// Error in exportObjects\n")
			printException(1, "while exporting objects")
		self.file.close()	

	
	# deprecated
	def writeNMesh_grreal(self, nObj, suffix):
		self.file.write("#define GRreal(n)\t(0x1c38+(n))\n")	
		self.file.write("GOURAUDTBL\tgrreal_%s[] = {\n" % (suffix))
		for obj in nObj:
			try:
				nmesh = obj.getData()
				if(type(nmesh) == Types.NMeshType):
					for f in nmesh.faces:
						if(f.smooth != 0):
							try:	
								self.file.write("\t{C_RGB(%s, %s, %s), C_RGB(%s, %s, %s), C_RGB(%s, %s, %s), C_RGB(%s, %s, %s)},\n" % (f.col[0][0], f.col[0][1], f.col[0][2], f.col[1][0], f.col[1][1], f.col[1][2], f.col[2][0], f.col[2][1], f.col[2][2], f.col[3][0], f.col[3][1], f.col[3][2]))
							except:
								self.file.write("\t{C_RGB(31, 31, 31), C_RGB(31, 31, 31), C_RGB(31, 31, 31), C_RGB(31, 31, 31)}, // ERROR: in writeNMesh_grreal\n");	
			except:
				self.file.write("// Error in writeNMesh_grreal\n")
				printException(1, "while exporting grreal")

		self.file.write("};\n\n")
		
	def writeNMesh_pntbl(self, nmesh, suffix):
		self.file.write("// Number of vertices: %s\n" % (len(nmesh.verts)))
		self.file.write("POINT\tpoint_%s[] = {\n" % (suffix))
		for v in nmesh.verts:
			try:
#				self.file.write("\tPOStoFIXED(%s, %s, %s),\n" % (v.co[0], v.co[1], v.co[2]))
				self.file.write("\tPOStoFIXED(%s, %s, %s),\n" % (v.co[0], -v.co[2], v.co[1]))
			except:
				self.file.write("\tPOStoFIXED(0.0, 0.0, 0.0), // ERROR: in writeNMesh_pntbl\n");
				printException(1, "while writing vertex coordinates")
		self.file.write("};\n\n")
	
	def writeNMesh_pltbl(self, nmesh, suffix):
		self.file.write("// Number of faces: %s\n" % (len(nmesh.faces)))
		self.file.write("POLYGON\tpolygon_%s[] = {\n" % (suffix))
		for f in nmesh.faces:
			try:
				# support faces with 3 vertices only
				if (len(f.v) == 4):
					lastIndex = 3
				else:
					lastIndex = 2
#				self.file.write("\tNORMAL(%s, %s, %s), VERTICES( %s, %s, %s, %s),\n" % (f.normal[0], f.normal[1], f.normal[2], f.v[0].index, f.v[1].index, f.v[2].index, f.v[3].index))
				self.file.write("\tNORMAL(%s, %s, %s), VERTICES( %s, %s, %s, %s),\n" % (f.normal[0], -f.normal[2], f.normal[1], f.v[0].index, f.v[1].index, f.v[2].index, f.v[lastIndex].index))
			except:
				self.file.write("\tNORMAL(0.0, 0.0, 0.0), VERTICES(0, 0, 0, 0), // ERROR: in writeNMesh_pltbl\n");	
		self.file.write("};\n\n")

	def writeNMesh_vntbl(self, nmesh, suffix):
		# to calculate normals (I don't know if user can set vertex normals manually, then this might overwrite this):
#		nmesh.update(recalc_normals=0, store_edges=0, vertex_shade=0)
#		nmesh.update(1, 0, 0)
		self.file.write("// Number of vertices: %s\n" % (len(nmesh.verts)))
		self.file.write("VECTOR\tnormal_%s[] = {\n" % (suffix))
		for v in nmesh.verts:
#			print "Vertex normal: %s" %(v.no)
			# is there a vertex normal defined for this face?
			# there might be a better way to test for presence of any vertex normal
			got_vn = 1
#			try:
#				print "test access: POStoFIXED(%s, %s, %s),\n" % (v.no[0], -v.no[2], v.no[1])
#			except:
#				got_vn = 0
				
			# if a vertex normal is not defined, let's compute it right here (not yet, blender might have computed it)
			if (got_vn == 1):
#				self.file.write("\tPOStoFIXED(%s, %s, %s),\n" % (v.no[0], v.no[1], v.no[2]))
				self.file.write("\tPOStoFIXED(%s, %s, %s),\n" % (v.no[0], -v.no[2], v.no[1]))
			else:
				self.file.write("\tPOStoFIXED(0.0, 0.0, 0.0), // ERROR: in writeNMesh_vntbl\n");
				printException(1, "while writing vertex normals")


		self.file.write("};\n\n")

	# need the obj for it's material list (although NMesh got it's own)
	def writeNMesh_attbl(self, obj, nmesh, suffix):
		# keep track if there are faces that use real time gouraud, 
		# if that's the case, we need to create a vertex normal table
		nGrRealFaces = 0
		self.file.write("// Number of faces: %s\n" % (len(nmesh.faces)))
		self.file.write("ATTR\tattribute_%s[] = {\n" % (suffix))
		for f in nmesh.faces:				
			try:
				# get the material for this face, try NMesh.GetMaterials(), then Object.GetMaterials(), last use dummy material
				try:
					mlist = nmesh.getMaterials()
#					print "len(nmesh.getMaterials()):%s, f.materialIndex: %s" %(len(mlist), f.materialIndex)
					if ( (len(mlist) > 0) and (f.materialIndex < len(mlist)) ):
						# fine, it got a material (in it's NMesh material list)
						m = mlist[f.materialIndex]
					else:
						mlist = obj.getMaterials()
#						print "len(obj.getMaterials()):%s, f.materialIndex: %s" %(len(mlist), f.materialIndex)
						if ( (len(mlist) > 0) and (f.materialIndex < len(mlist)) ):
							# fine, it got a material (in it's Object material list)
							m = mlist[f.materialIndex]
						else:
							# No material! Use a global dummy material, check if it already exists
							print "This face got no material!"
							try:
								m = Material.Get('sglDummyMaterial')
#								print "Will use sglDummyMaterial..."
							except:
	#							print "Will create sglDummyMaterial for the first time..."
								m = Material.New('sglDummyMaterial')
								m.rgbCol = [0.5, 0.5, 0.5]  
				except:
					printException(1, "while getting a material for this face")

				# testing texture export
				usesTexture = []
				try:
					tlist = m.getTextures()
					for t in tlist:
						if (t != None):
#							print "TeCo: %s" % (t.texco)
#							if (t.tex.getType() != 'Image'):
							if (t.tex.type == Texture.Types.IMAGE):							
								if ((t.mapto & Texture.MapTo.COL) != 0):
									if (f.uv != []):
										if (t.texco != Texture.TexCo.UV):
											print "WARNING: this texture is not mapped using it's UV coordinates.\nIt will be exported, but may look different on Saturn."
#										print "Exporting image texture...uv: %s" % (f.uv)
										im = t.tex.getImage()
	#									print "Image of t: %s, of f: %s" %(im, f.image)
										if (usesTexture == []):
											usesTexture = self.addTexture(im, f.uv)
										else:
											print "WARNING: This face got multiple textures assigned, only the first one has been exported!"
	#									self.test(im, f.uv, t.tex.getName(), self.texIndex) #"name"
	#									self.texIndex += 1
	
									else:
										print "ERROR: can't export texture, no UV coordinates assigned!"
#								else:
#									print "Not a normal color texture!"

#							else:
#								print "Not an image texture, but: %s" % (t.tex.getType())

#						else:
#							print "Texture channel empty!"
				except:
					printException(1, "while adding texture to dictionary")
					
				self.file.write("\tATTRIBUTE(")
				# plane
					# specify:
					# either Single_Plane or Dual_Plane
				if (f.mode & NMesh.FaceModes.TWOSIDE) == NMesh.FaceModes.TWOSIDE:
					self.file.write("Dual_Plane, ")
				else:
					self.file.write("Single_Plane, ")
				# sort
					# specify:
					# either SORT_MIN, SORT_CEN, SORT_MAX or SORT_BFR
				self.file.write("SORT_CEN, ")
				# texture
					# specify:
					# either No_Texture or texture table index
				if (usesTexture != []):	#(f.mode & NMesh.FaceModes.TEX) == NMesh.FaceModes.TEX:
					# got texture, specify:
					# texture id
					self.file.write("%s, " % usesTexture[2])
				else:
					# no texture, specify:
					# No_Texture
					self.file.write("No_Texture, ")
				# color
					# specify:
					# either polygon RGB color
					# or palete offset or No_Palet
				if (usesTexture != []):	#f.mode & NMesh.FaceModes.TEX) == NMesh.FaceModes.TEX:
					# got texture, specify:
					# No_Palet or palete offset
					self.file.write("No_Palet, ")
				else:
					# no texture, specify:
					# RGB face color
					try:
#						rgbCol = nmesh.getMaterials()[f.materialIndex].rgbCol
						rgbCol = m.rgbCol
						self.file.write("C_RGB(%s, %s, %s), " % (int(31*rgbCol[0]), int(31*rgbCol[1]), int(31*rgbCol[2])))
					except:
						self.file.write("C_RGB(16, 16, 16), ")
						printException(0, "while writing polygon color")
				# gouraud
				usesGouraudType = ""
					# beware: vertex painting works with smooth = 0, too!
				if ((m.mode & Material.Modes.VCOL_PAINT) !=  0):
					# vertex painted (gouraud!)
#					print "Vertex paint!"
					if(f.smooth == 0):
						usesGouraudType = "static"
						self.file.write("GrStatic(%s), " % (len(self.GrStaticTbl)))
						self.GrStaticTbl.append(f.col)
					else:
						usesGouraudType = "real"
						self.file.write("GrReal(%s), " % (len(self.GrRealTbl)))
						self.GrRealTbl.append(f.col)
				else:
					# No vertex paint (gouraud or flat)
#					print "No vertex paint!"
					if(f.smooth == 0):
						self.file.write("No_Gouraud, ")
					else:
						usesGouraudType = "real"
						# where to take the vertex colors from?
						self.file.write("GrReal(%s), " % (len(self.GrRealTbl)))
						try:
							vcl = []
							vcl.append(NMesh.Col(16 << 3, 16 << 3, 16 << 3, 16 << 3))
							vcl.append(NMesh.Col(16 << 3, 16 << 3, 16 << 3, 16 << 3))
							vcl.append(NMesh.Col(16 << 3, 16 << 3, 16 << 3, 16 << 3))
							vcl.append(NMesh.Col(16 << 3, 16 << 3, 16 << 3, 16 << 3))
							self.GrRealTbl.append(vcl)
						except:
							printException(1, "while creating vertex color list")
				# mode
					# specify:
					# No_Window, Window_In, Window_Out
					# MESHon, MESHoff
					# ECdis, ECenb
					# SPdis, SPenb
					# CL16Bnk, CL16Look, CL64Bnk, CL128Bnk, CL256Bnk, CL32KRGB
					# CL_Replace, CL_Shadow, CL_Half, CL_Trans, CL_Gouraud
				self.file.write("No_Window")
				self.file.write("|MESHoff")
				self.file.write("|ECenb")
				self.file.write("|SPdis")
				self.file.write("|CL32KRGB")
				if(usesGouraudType == ""):
					self.file.write("|CL_Replace")
				else:
					self.file.write("|CL_Gouraud")

				self.file.write(", ")
				# dir
					# specify:
					# sprNoflip, sprHflip, sprVflip, sprHVflip, 
					# sprPolygon, sprLine, sprPolyline
				if (usesTexture != []):	#f.mode & NMesh.FaceModes.TEX) == NMesh.FaceModes.TEX:
					# got texture, specify either:
					# sprNoflip, sprHflip, sprVflip or sprHVflip
					if (usesTexture[0] == 0):
						if (usesTexture[1] == 0):
							self.file.write("sprNoflip, ")
						else:
							self.file.write("sprVflip, ")
					else:
						if (usesTexture[1] == 0):
							self.file.write("sprHflip, ")
						else:
							self.file.write("sprHVflip, ")
				else:
					# no texture, specify either
					# sprPolygon, sprPolyline or sprLine
					self.file.write("sprPolygon, ")
				# option
					# specify:
					# UseLight, UseClip, UsePalete
					# UseGouraud, UseDepth, UseNear <- SGL 3 add
					# or No_Option
				if(usesGouraudType == "real"):
					self.file.write("UseGouraud")
					# here I don't need to update it at multiple locations
					nGrRealFaces += 1
				else:
					self.file.write("No_Option")
				
				self.file.write("),\n")
			except:
				self.file.write("// ERROR: in writeNMesh_altbl\n");	
		self.file.write("};\n\n")
		return nGrRealFaces
					
	def writeNMesh_XPDATA(self, nmesh, suffix, got_vntbl):
		try:
			self.file.write("XPDATA\txpdata_%s[] = {\n" % (suffix))
			self.file.write("\tpoint_%s, sizeof(point_%s)/sizeof(POINT),\n" % (suffix, suffix))
			self.file.write("\tpolygon_%s, sizeof(polygon_%s)/sizeof(POLYGON),\n" % (suffix, suffix))
			self.file.write("\tattribute_%s,\n" % (suffix))
			if(got_vntbl == 0):
				self.file.write("\tNULL,\n")
			else:
				self.file.write("\tnormal_%s,\n" % (suffix))
			self.file.write("};\n\n")
		except:
			self.file.write("// ERROR: in writeNMesh_XPDATA\n")
			printException(1, "while writing XPDATA")

	def writeNMesh_OBJECT(self, obj, suffix_obj, suffix_dat, suffix_child, suffix_sibling):
		try:
			self.file.write("OBJECT\tobject_%s[] = {\n" % (suffix_obj))
			if(suffix_dat == ""):
				self.file.write("\t(PDATA *)NULL,\n")
			else:
				self.file.write("\t(PDATA *)xpdata_%s,\n" % (suffix_dat))
			pos = obj.getLocation()
#			self.file.write("\tTRANSLATION(%s, %s, %s),\n" % (pos[0], pos[1], pos[2]))
			self.file.write("\tTRANSLATION(%s, %s, %s),\n" % (pos[0], -pos[2], pos[1]))
			rot = obj.getEuler()
#			self.file.write("\tROTATION(%s, %s, %s),\n" % (rot[0], rot[1], rot[2]))
#			self.file.write("\tROTATION(%s, %s, %s),\n" % (rot[0], -rot[2], rot[1]))

			my_pi = 3.141592654
			self.file.write("\tROTATION(%s, %s, %s),\n" % (180*rot[0]/my_pi, -180*rot[2]/my_pi, 180*rot[1]/my_pi))
# math not available
#			self.file.write("\tROTATION(%s, %s, %s),\n" % (degrees(rot[0]), degrees(-rot[2]), degrees(rot[1]) ))
			scl = obj.getSize()
#			self.file.write("\tSCALING(%s, %s, %s),\n" % (scl[0], scl[1], scl[2]))
			self.file.write("\tSCALING(%s, %s, %s),\n" % (scl[0], scl[2], scl[1]))
			if (suffix_child == ""):
				self.file.write("\tNULL,\t// Got no children\n")
			else:
				self.file.write("\tobject_%s,\t// This is the last child exported\n" % (suffix_child))
			if (suffix_sibling == ""):
				self.file.write("\tNULL,\t// Got no previous sibling(this is the first)\n")
			else:
				self.file.write("\tobject_%s,\t// This is the previous sibling exported\n" % (suffix_sibling))
			self.file.write("};\n\n")
		except:
			self.file.write("// ERROR: in writeNMesh_OBJECT\n")	
			printException(1, "while writing OBJECT")
	
	def writeObject(self, obj, name_child, name_sibling):
		# only export SELECTED mesh data objects
		self.file.write("// Exporting object %s\n" % obj.getName())
		if(self.SelectedObjectList.count(obj) == 1 and type(obj.getData()) == Types.NMeshType):
			# avoid converting linked duplicates to full duplicates
			# object data is named using data name
			suffix = obj.getData(1).replace('.', '_')
			if (self.DataNameList.count(obj.getData(1)) == 0):
				# the data hasn't been written before
				self.DataNameList.append(obj.getData(1))
				nmesh = obj.getData()
			
				self.writeNMesh_pntbl(nmesh, suffix)	
				self.writeNMesh_pltbl(nmesh, suffix)
				nGrRealFaces = self.writeNMesh_attbl(obj, nmesh, suffix)
				if (nGrRealFaces > 0):
					self.writeNMesh_vntbl(nmesh, suffix)
				self.writeNMesh_XPDATA(nmesh, suffix, nGrRealFaces)
				self.file.write("// This Object is NOT a linked copy!\n")
			else:
				self.file.write("// This Object is a linked copy!\n")
		else:
#			suffix = "empty"
			suffix = ""
			
		# objects are named with object name	
		self.writeNMesh_OBJECT(obj, obj.getName().replace('.', '_'), suffix, name_child.replace('.', '_'), name_sibling.replace('.', '_'))
	
	def writeHeader(self):
		self.file.write("/* Generated with:\n * SegaSaturn.py - python script for blender to export to SGL format.\n * by The Rockin'-B, http://www.rockin-b.de\n */\n\n")
		# include gouraud tables, (relative path name)
		if (self.filename_gr.count('/') != 0):
			lst = self.filename_gr.split('/')
#			print "lst: %s" % (lst)
#			print "basename: %s" % (lst[len(lst)-1])
			self.file.write("#include \"%s\"\n\n" % (lst[len(lst)-1]))
		elif (self.filename_gr.count("\\") != 0):
			lst = self.filename_gr.split("\\")
#			print "lst: %s" % (lst)
#			print "basename: %s" % (lst[len(lst)-1])
			self.file.write("#include \"%s\"\n\n" % (lst[len(lst)-1]))
		else:
#			print "No seperator in: %s" % (self.filename_gr)
			self.file.write("#include \"%s\"\n\n" % (self.filename_gr))

#		self.file.write("//The empty xpdata object is used for empty root objects (instead of a NULL pointer reference)\n")
#		self.file.write("XPDATA\txpdata_empty[] = {\n\tNULL, 0,\n\tNULL, 0,\n\tNULL,\n\tNULL,\n};\n")

	# note: abs is just a workaround, because uv can be negative
	# WRONG: assumes rectangular uv coordinates
	def sampleTexture(self, im, uv, w, h):
		PixelList = []
#		print "w: %s, h: %s" % (w, h)
		sz = im.getSize()
#		print "img_w: %s, img_h: %s" % (sz[0], sz[1])
#		print "uv: %s" % (uv)

		# map uv to pixel space
			# (top) left side
		x_left = uv[0][0]*im.getMaxXY()[0]
		y_left = uv[0][1]*im.getMaxXY()[1]
		# left side, one step down
		x_left_add = (uv[3][0]*im.getMaxXY()[0] - x_left) / (w-1)
		y_left_add = (uv[3][1]*im.getMaxXY()[1] - y_left)
		if (h > 1):
			y_left_add /= (h-1)
			# (top) right side
		x_right = uv[1][0]*im.getMaxXY()[0]
		y_right = uv[1][1]*im.getMaxXY()[1]
		# right side, one step down
		x_right_add = (uv[2][0]*im.getMaxXY()[0] - x_right) / (w-1)
		y_right_add = (uv[2][1]*im.getMaxXY()[1] - y_right)
		if (h > 1):
			y_right_add /= (h-1)
		# to ensure pixel coordinates are inside the image
		x_mask = -int(sz[0])
		y_mask = -int(sz[1])
#		print "sz[0] %x, x_mask %x" % (sz[0], x_mask)

		# ensure it's in image space
#		x = abs(x % sz[0])
#		y = abs(y % sz[1])
#		print "Here 2"
			
#		print "Here 3"
		
		for r in range(h):
			x = x_left
			y = y_left
			x_add = (x_right - x_left) / (w-1)
			y_add = (y_right - y_left) / (w-1)
			for c in range(w):
#				print "Pixel (x,y): (%s, %s)" % (x, y)
				if (self.bilinear):
					# 4 pixels around, get the coordinates of the topleft one
						# ensure x,y are in image space (slow version, for every pixel)
					xl = int(x-0.5) % sz[0]
					if (xl < 0):
						xl += sz[0]
					yt = int(y-0.5) % sz[1]
					if (yt < 0):
						yt += sz[1]
					# fetch the four pixels
						# stay inside image (the first one already is, for sure)
					xr = xl+1
					if (xr >= sz[0]):
						xr = 0
					yb = yt+1
					if (yb >= sz[1]):
						yb = 0
						# fetch
					pxtl = im.getPixelF(xl, yt)
					pxtr = im.getPixelF(xr, yt)
					pxbl = im.getPixelF(xl, yb)
					pxbr = im.getPixelF(xr, yb)
					# get x,y fraction relative to topleft pixel
					x_fract = x % 1
					if (x_fract < 0):
						x_fract = 1-x_fract
					y_fract = y % 1
					if (y_fract < 0):
						y_fract = 1-y_fract
					# interpolate
					px = im.getPixelF(xl, yt)
					for i in range(4):
						cml = pxtl[i] + y_fract*(pxbl[i] - pxtl[i])
						cmr = pxtr[i] + y_fract*(pxbr[i] - pxtr[i])
						px[i] = cml + x_fract*(cmr - cml)
				else:
					# nearest neighbor
						# ensure x,y are in image space (slow version, for every pixel)
					x2 = int(x+0.5) % sz[0]
					if (x2 < 0):
						x2 += sz[0]
					y2 = int(y+0.5) % sz[1]
					if (y2 < 0):
						y2 += sz[1]
					px = im.getPixelF(x2, y2)
	#				print "x %s, y %s, x2 %s, y2 %s, iw %s, ih %s" % (x, y, x2, y2, sz[0], sz[1])
#					try:
#						PixelList.append(im.getPixelF(x2, y2))
#					except:
#						print "Texel Error: x %s, y %s, x2 %s, y2 %s, iw %s, ih %s" % (x, y, x2, y2, sz[0], sz[1])

				PixelList.append(px)
				
				x += x_add
				y += y_add
#				if (x < 0):
#					x += sz[0]
#				elif (x >= sz[0]):
#					x -= sz[0]
			
#			if (y < 0):
#				y += sz[1]
#			elif (y >= sz[1]):
#				y -= sz[1]
			x_left += x_left_add
			y_left += y_left_add
			x_right += x_right_add
			y_right += y_right_add
			
		return PixelList

	def writeTexture(self, pxlst, suffix):
		self.file.write("Uint16\ttexture_%s[] = {" % (suffix))
		c = 0
		for px in pxlst:
			# manage linebreak
			if (c == 0):
				self.file.write("\n\t")
				c = 3
			else:
				c -= 1
			self.file.write("C_RGB(%s, %s, %s), " % (int(31*px[0]), int(31*px[1]), int(31*px[2])) )

		self.file.write("\n};\n\n")
		
	def uvFrom0To1(self, uv):
		# transform to interval 0..1
		my_uv = [uv[0]  % 1, uv[1]  % 1]
		if(my_uv[0] < 0):
			my_uv[0] += 1
		if(my_uv[1] < 0):
			my_uv[1] += 1
		return my_uv
		
	# transform to value interval 0...1 and return 1 if the coordinates are below a certain threshold
	# don't expect floats to be equal, even with a small difference, the resulting integer image coordinate is the same
	def uvEquals(self, uva, uvb):
		# because function parameters don't support item assignment (or something like that...)
#		uva = self.uvFrom0To1(uv_a)
#		uvb = self.uvFrom0To1(uv_b)
		# test threshold
		if ( (abs(uva[0]-uvb[0]) < self.uvThreshold) and (abs(uva[1]-uvb[1]) < self.uvThreshold) ):
			return 1
		else:
			return 0
	
	# returns (hflip, vflip, texIndex)
	# if len(uv) < 4 will be extended to length 4
	# images assumed to have a unique name
	def addTexture(self, im, uv):
		if (len(uv) == 0):
			print "ERROR: the texture could not be added, because it got no UV coordinates assigned!"
			return []
		# ensure uv has length 4
		if (len(uv) < 4):
			uvlast = uv[len(uv)-1]
			for i in range(4-len(uv)):
				uv.append(uvlast)
		# each texture/image got a list of smaller subtextures to be exported
		if (self.TextureDict.has_key(im.getName())):
			SubTexLst = self.TextureDict[im.getName()][1]
		else:
			SubTexLst = []
#		print "Here 3: let(SubTexLst): %s" % (len(SubTexLst))
		# search if the subtexture is already in the list (maybe h/v flipped)
		for st in SubTexLst:
#			print "st: %s" % (st)
			uvlst = st[1]
#			print "Here 3.1, uvl: %s" % (uv)
#			print "Here 3.1, uvlst: %s" % (uvlst)
			# uv reordering through h/v flipping
			if (self.uvEquals(uv[0], uvlst[0]) and self.uvEquals(uv[1], uvlst[1]) and self.uvEquals(uv[2], uvlst[2]) and self.uvEquals(uv[3], uvlst[3])):
				# 0123 noflip (clockwise, starting topleft)
				return 0,0,st[0]
			elif (self.uvEquals(uv[0], uvlst[1]) and self.uvEquals(uv[1], uvlst[0]) and self.uvEquals(uv[2], uvlst[3]) and self.uvEquals(uv[3], uvlst[2])):
				# 1032 hflip
				return 1,0,st[0]
			elif (self.uvEquals(uv[0], uvlst[3]) and self.uvEquals(uv[1], uvlst[2]) and self.uvEquals(uv[2], uvlst[1]) and self.uvEquals(uv[3], uvlst[0])):
				# 3210 vflip
				return 0,1,st[0]
			elif (self.uvEquals(uv[0], uvlst[2]) and self.uvEquals(uv[1], uvlst[3]) and self.uvEquals(uv[2], uvlst[0]) and self.uvEquals(uv[3], uvlst[1])):
				# 2301 hvflip
				return 1,1,st[0]
		# not already in the list, it's the first
			# WARNING: this wouldn't be a flipping, but a clockwise shift of coordinate indices, can't do it now
			# assign initial flipping intelligently, let longest side be width
#		w = (uvFrom0To1(uv[0])[0] - uvFrom0To1(uv[1])[0])**2 + (uvFrom0To1(uv[0])[1] - uvFrom0To1(uv[1])[1])**2
#		w += (uvFrom0To1(uv[3])[0] - uvFrom0To1(uv[2])[0])**2 + (uvFrom0To1(uv[3])[1] - uvFrom0To1(uv[2])[1])**2
#		w /= 2
#		h = (uvFrom0To1(uv[0])[0] - uvFrom0To1(uv[3])[0])**2 + (uvFrom0To1(uv[0])[1] - uvFrom0To1(uv[3])[1])**2
#		h += (uvFrom0To1(uv[1])[0] - uvFrom0To1(uv[2])[0])**2 + (uvFrom0To1(uv[1])[1] - uvFrom0To1(uv[2])[1])**2
#		h /= 2
#		if (w < h)
		# write back to dict
			# the correct data structure format can be seen here
		w = self.uvComputeWeight(uv)
		self.texWeightSum += w
		SubTexLst.append([self.texIndex, uv, w])
		self.TextureDict[im.getName()] = [im, SubTexLst]
		self.texIndex += 1
		return 0,0,self.texIndex-1
	
	def test(self, im, uv, name, num):
		print "TEST: sampleTexture..."
		PixelList = self.sampleTexture(im, uv, 16, 8)
		self.writeTexture(PixelList, name, num)
#		for i in PixelList:
#			print "pixel %f, %f, %f" % (i[0], i[1], i[2])
		print "TEST: finished!"

	def uvSquareLength(self, uva, uvb):
		return (uvb[0]-uva[0])**2 + (uvb[1]-uva[1])**2
	
	# this weight should be closely related to the area consumed
	def uvComputeWeight(self, uvlst4):
		# see image on german wikipedia entry for Viereck
		# uv points from 0 to 3 refer to Viereck points: A, D, C, B
		d2 = self.uvSquareLength(uvlst4[0], uvlst4[1])
		c2 = self.uvSquareLength(uvlst4[1], uvlst4[2])
		b2 = self.uvSquareLength(uvlst4[2], uvlst4[3])
		a2 = self.uvSquareLength(uvlst4[3], uvlst4[0])

		e2 = self.uvSquareLength(uvlst4[0], uvlst4[2])
		f2 = self.uvSquareLength(uvlst4[1], uvlst4[3])
		
		A = 0.25 * sqrt(4*e2*f2 - (b2+d2-a2-c2)**2)
		return A

	# old, unused
	def uvComputeWeightSurround(self, uvlst4):
		# the summed length of the edges, no need to sqrt it
		w = 0
		for i in range(4):
#			print "i is %s, l is %s" % (i, len(uvlst4))
			j = i+1
			if (j == 4):
				j = 0
			w += (uvlst4[i][0] - uvlst4[j][0])**2 + (uvlst4[i][1] - uvlst4[j][1])**2
		return w
	
	# + cares about width/height restrictions
	# - doesn't take poylgon shape into account
	# (old, unused version had factor 1)
	# (+ when w_factor = 1, ensures that width is never smaller than height)
	def getWidthHeightFactor(self, my_max_pixnum, w_start, h_start, w_max, h_max, w_factor):
		w = w_start
		h = h_start
		h_last = 0
		w_last = 0
		while (w*h <= my_max_pixnum):
			h_last = 0
			w_last = 0
#			print "w %s, h %s" % (w, h)
			if (h < w*w_factor and h <= (h_max-1) ):
				h += 1
				h_last = 1
			elif (w <= (w_max-8) ):
				w += 8
				w_last = 8
			elif (h <= (h_max-1) ):	# the ratio cannot be fullfilled, but try to use granted pixnum nevertheless
				h += 1
				h_last = 1
			else:
				break
		# undo the last step, which (possibly) exceeded the size
#		print "h %s, w %s, w_factored %s, h_last %s, w_last %s" % (h, w, w*w_factor, h_last, w_last)
		h -= h_last
		w -= w_last
		return w,h

	
	def getWidthHeight(self, im, my_max_pixnum, w_start, h_start, w_max, h_max, uvlst4):
		# determine a width/height ratio
			# width and height, mean
		w_m = sqrt(self.uvSquareLength(uvlst4[0], uvlst4[1])) + sqrt(self.uvSquareLength(uvlst4[2], uvlst4[3]))
		h_m = sqrt(self.uvSquareLength(uvlst4[1], uvlst4[2])) + sqrt(self.uvSquareLength(uvlst4[3], uvlst4[0]))
			# width and height should always approximate this ratio:
			# h_m / w_m2 == height/width
		w_factor = h_m / w_m
#		print "width top %s, bottom %s" % (sqrt(self.uvSquareLength(uvlst4[0], uvlst4[1])), sqrt(self.uvSquareLength(uvlst4[2], uvlst4[3])))
#		print "height left %s, right %s" % (sqrt(self.uvSquareLength(uvlst4[3], uvlst4[0])), sqrt(self.uvSquareLength(uvlst4[1], uvlst4[2])))
#		print "h_m %s, w_m %s, w_factor %s" % (h_m, w_m, w_factor)

		# avoid wasting texture space by oversampling the image
		# so further restrict given w_max and h_max to the actual image dimensions
		# again use the middle distance of PIXELS
		sz = im.getSize()
		wp_m = sqrt(((uvlst4[1][0] - uvlst4[0][0])*sz[0])**2 + ((uvlst4[1][1] - uvlst4[0][1])*sz[1])**2)
		wp_m += sqrt(((uvlst4[3][0] - uvlst4[2][0])*sz[0])**2 + ((uvlst4[3][1] - uvlst4[2][1])*sz[1])**2)
		wp_m /= 2

		hp_m = sqrt(((uvlst4[2][0] - uvlst4[1][0])*sz[0])**2 + ((uvlst4[2][1] - uvlst4[1][1])*sz[1])**2)
		hp_m += sqrt(((uvlst4[0][0] - uvlst4[3][0])*sz[0])**2 + ((uvlst4[0][1] - uvlst4[3][1])*sz[1])**2)
		hp_m /= 2
		
		if (wp_m < w_max):
			print "Note: max width %s is restricted to %s, to avoid wasting texture space by oversampling" % (w_max, wp_m)
			w_max = wp_m
		if (hp_m < h_max):
			print "Note: max height %s is restricted to %s, to avoid wasting texture space by oversampling" % (h_max, hp_m)
			h_max = hp_m
#		print "wp_m %s, hp_m %s" % (wp_m, hp_m)
		# approximate my_max_pix_num while keeping the ratio
		return self.getWidthHeightFactor(my_max_pixnum, w_start, h_start, w_max, h_max, w_factor)
	
	# loops over all subtextures to export,
	# keeps track of that subtexture, that got the most pixels granted in relation to the other subtextures
	def computeWidthsHeightsOnePass(self, pixnum, w_max, h_max, updateWeightSum):
		max_idx = -1
		max_pixPerWeight = 0
		
		texWeightSum = self.texWeightSum
		for imgname in self.TextureDict.keys():
			im = self.TextureDict[imgname][0]
			SubTexLst = self.TextureDict[imgname][1]
			for st in SubTexLst:
				tex = self.texTbl[st[0]]
				# determine size
					# how many pixels can this texture consume?
				my_max_pixnum = pixnum * st[2]
				if (texWeightSum != 0):
					my_max_pixnum /= texWeightSum
				my_max_pixnum += tex[0]*tex[1]
#				print "my_weight %s, sum %s" % (st[2], texWeightSum)
#				print "my_max_pixnum %s" % (my_max_pixnum)
				print "id %s, w_start %s, h_start %s, pix %s" % (st[0], tex[0], tex[1], my_max_pixnum)
					# approximate max_pixnum
				(w, h) = self.getWidthHeight(im, my_max_pixnum, tex[0], tex[1], w_max, h_max, st[1])
				print "\tw %s, h %s" % (w, h)
				# save dimensions for TEXTURE table and name for PICTURE table
				self.texTbl[st[0]] = [w, h, tex[2]]
				# update loop
				pixnum -= w*h - tex[0]*tex[1]
				if (updateWeightSum):
					texWeightSum -= st[2]
				# update search for biggest texture
#				print "max_pixPerWeight %s" % (w*h/st[2])
				if (max_pixPerWeight <= w*h/st[2]):
					max_idx = st[0]
#					print "new max_idx %s" % (max_idx)
					max_pixPerWeight = w*h/st[2]
		return max_idx, pixnum


	# multi pass
	def computeWidthsHeights(self, pixnum, w_max, h_max):
		# initialize texTbl for first pass, set the start widths and heights to the minimum
		self.texTbl = {}
		for tidx in range(self.texIndex):
			self.texTbl[tidx] = [8, 1, ""]
		pixnum -= self.texIndex*8*1
		# enough pixels?
		if (pixnum < 0):
			return

		# use olf method until a better one is finished
		self.computeWidthsHeightsOnePass(pixnum, w_max, h_max, 1)
		return

		# distribute number of pixels evenly among the subtextures
		(max_idx, pixnum)  = self.computeWidthsHeightsOnePass(pixnum, w_max, h_max, 0)
		# there is some space left, try
		finished = 0
		while(not finished):
			# allow one subtexture to be increased
			# choose the one which has got too few space for it's size, compared to the others
			
			finished = 1

		# old multi-pass approach, no good
		for i in range(2):
			print "pass %i" % (i)
			(max_idx, pixnum)  = self.computeWidthsHeightsOnePass(pixnum, w_max, h_max, 1)
			print "remaining pixels %s, max_idx %s" % (pixnum, max_idx)
			# take something away from the biggest texture
			if (max_idx >= 0):
					# rather dumb method
				tex = self.texTbl[max_idx]
				w = tex[0]
				h = tex[1]
				if (w > h and h > 1):
					h -= 1
				elif (w <= h and w > 8):
					w -= 8
				else:
					print "Couldn't make biggest texture smaller"
				pixnum += tex[0]*tex[1] - w*h
				tex[0] = w
				tex[1] = h
				self.texTbl[max_idx] = tex
			else:
				print "No biggest subtexture found"
		
		
		
		
	# all textures to export have been put in the dictionary before
	# Use the complete granted texture space (number of pixels) intelligently,
	# by asigning each texture a weight which determines the amount of pixels that it can consume.
	# Because texture width is a multiple of eight, in most cases a texture cannot consume exactly as much pixels as granted,
	# so keep track of the actual pixel number written.

	# character size:
	# horizontal: 8-504 pixels (8-pixel units),
	# vertical:   1-255
	def exportTextures(self, pixnum, w_max, h_max):
		self.file = open(self.filename_txr, "w")
		self.computeWidthsHeights(pixnum, w_max, h_max)
		texTbl = {}
#		print "Here 1"
		texturesProcessed = 0
		for imgname in self.TextureDict.keys():
			SubTexLst = self.TextureDict[imgname][1]
#			print "Here 2"
			texnum = 0
			for st in SubTexLst:
#				print "Here 3"
				# determine size
					# how many pixels can this texture consume?
#				my_max_pixnum = pixnum * st[2]/self.texWeightSum
#				print "my_weight %s, sum %s" % (st[2], self.texWeightSum)
#				print "my_max_pixnum %s" % (my_max_pixnum)
					# approximate max_pixnum
#				(w, h) = self.getWidthHeight(my_max_pixnum, 8, 1, w_max, h_max, st[1])
				# export texture
				tex = self.texTbl[st[0]]
				w = tex[0]
				h = tex[1]
				Window.DrawProgressBar(0.3 + 0.7*texturesProcessed / self.texIndex, "Generating textures...")
				pxlst = self.sampleTexture(self.TextureDict[imgname][0], st[1], w, h)
#				print "Here 4"
				self.file.write("// width = %s,\t\theight = %s\n" % (w, h))
				suffix = "%s_%s" % (imgname.replace('.', '_'), texnum)
				self.writeTexture(pxlst, suffix)
#				print "Here 5"
				# save dimensions for TEXTURE table and name for PICTURE table
				tex[2] = suffix
				texTbl[st[0]] = tex	#[w, h, suffix]

				# update loop
				pixnum -= w*h
				texnum += 1
				self.texWeightSum -= st[2]
				texturesProcessed += 1

#		print "Here 6"
		# export TEXTURE table
		self.file.write("TEXTURE\tTexTbl[] = {\n")
		offset = 0
#		print "texTbl.keys() %s" % (texTbl.keys())
#		print "Here a1"
		keys = texTbl.keys()
		keys.sort()
		for k in keys:
			sz2 = texTbl[k]
#			print "Here a2"
			self.file.write("\tTEXTBL(\t%s,\t%s, CGADDRESS+%s),\n" % (sz2[0], sz2[1], offset))
			offset += sz2[0]*sz2[1]*2
#			print "Here a3"
		self.file.write("};\n\n")
		
		# export PICTURE table
#		print "Here a"
		self.file.write("PICTURE\tPicTbl[] = {\n")
		for imgname in self.TextureDict.keys():
			SubTexLst = self.TextureDict[imgname][1]
			for st in SubTexLst:
#				print "Here b"
				self.file.write("\tPICTBL(\t%s, COL_32K, texture_%s),\n" % (st[0], texTbl[st[0]][2]))
#		print "Here c"
		self.file.write("};\n\n")
		self.file.write("#define\tPicTblLength\t%s\n" % (self.texIndex))
		
		
		self.file.write("// number of unused pixels = %s\n" % (pixnum))
		self.file.close()
		
	def write_gouraud_table(self, tbls, suffix):
		self.file.write("static GOURAUDTBL\tGr%sTbl[] = {\n" % (suffix))
#		print "Gouraud table length: %i" % (len(tbls))
		for tbl in tbls:
#			print "Color table length: %i" % (len(tbl))
			# if less then 4 colors in, complete the list with dummy stuff
			try:
				if (len(tbl) < 4):
					for i in range(4-len(tbl)):
						print "Appending item %s" % (i)
						try:
							tbl.append(NMesh.Col(255, 255, 255, 255))
						except:
							printException(1, "while creating a new MNCol object")
			except:
				printException(1, "while completing gouraud table color list")
#			print "Color table length2: %i" % (len(tbl))
			self.file.write("\t{")
			for col in tbl:
#				print "Color: %s" % (col)
				self.file.write("C_RGB(%s, %s, %s)," % (col.r >> 3, col.g >> 3, col.b >> 3))
			# if less then 4 colors, complete the list with dummy stuff
			self.file.write("},\n")
		self.file.write("};\n\n")

	def write_gouraud_tables(self):
		self.file = open(self.filename_gr, "w")

		self.write_gouraud_table(self.GrStaticTbl, "Static")
		self.write_gouraud_table(self.GrRealTbl, "Real")

		self.file.write("\n// These are VRAM relative addresses divided by eight, see ST-237-R1-051795.pdf, page 125\n")	
		self.file.write("#ifndef GrStart\n")	
		self.file.write("#define GrStart            0x1c38\n")	
		self.file.write("#endif\n")	
		self.file.write("#define GrStatic(n)        (GrStart + (n))\n")	
		self.file.write("#define GrReal(n)          (GrStart + (sizeof(GrStaticTbl)/sizeof(GOURAUDTBL)) + (n))\n")	
		self.file.write("\n// The address in VDP1 VRAM, where the static gouraud table is written to, once at startup\n")	
		self.file.write("#define GrStaticDmaStart   ((void *)(SpriteVRAM + (GrStatic(0)<<3) ))\n")	
		self.file.write("\n// not really needed, since slInitGouraud(.., GrReal(0),.) :\n")	
		self.file.write("#define GrRealDmaStart     ((void *)(SpriteVRAM + (GrReal(0)<<3) ))\n")	
		self.file.write("\n// Size of gouraud tables (number of faces using static or real time gouraud)\n")	
		self.file.write("#define GrStaticTblLength  %s\n" % (len(self.GrStaticTbl)))	
		self.file.write("#define GrRealTblLength    %s\n" % (len(self.GrRealTbl)))	

		self.file.write("\n// Work areas for real time gouraud\n")	
		self.file.write("static GOURAUDTBL          GrRealWorkTbl[GrRealTblLength];\n")
		self.file.write("static Uint8               GrRealWorkSpc[GrRealTblLength];\n")

		self.file.close()
		

def file_callback(filename):
	print "You selected file: ", filename
	global mdl
	# overwrite/update filenames to the choosen ones
	mdl.setFilenames(filename)
	
	# if something is selected, export selected only,
	# else export all
	objects = Blender.Object.GetSelected()
	if (len(objects)== 0):
		objects = Blender.Object.Get()
		print "No objects selected, exporting all objects\n"
	else:
		print "Exporting selected objects only.\nTo export all, either select all or select nothing!\n"
	
	Window.DrawProgressBar(0, "Exporting 3D objects...")
	mdl.exportObjects(objects)
	Window.DrawProgressBar(0.2, "Writing gouraud table...")
	mdl.write_gouraud_tables()
	mdl.exportTextures(mdl.maxTexels, mdl.maxWidth, mdl.maxHeight)#256*1024/2)
	Window.DrawProgressBar(1, "Finished")

	global nCriticalExceptions
	if (nCriticalExceptions > 0):
		print "There were %i critical exceptions, check the output file!" % (nCriticalExceptions)
	else:
		print "Finished without exceptions!"




def registry_save():
	d = {}
	d['maxWidth'] = mdl.maxWidth
	d['maxHeight'] = mdl.maxHeight
	d['maxTexels'] = mdl.maxTexels
	d['bilinear'] = mdl.bilinear
	# cache = True: data is also saved to a file
	Blender.Registry.SetKey('SegaSaturnSGL', d, True)


def registry_load():
	d = Registry.GetKey('SegaSaturnSGL', True) # True to check on disk also
	if d:
		try:
			mdl.maxWidth = d['maxWidth']
			mdl.maxHeight = d['maxHeight']
			mdl.maxTexels = d['maxTexels']
			mdl. bilinear = d['bilinear']
		except: registry_save() # if data isn't valid rewrite it

try:
	arg = __script__['arg']
except:
	printException(1, "Error: could not get script arguments!")
else:
#	fname = Blender.sys.makename(ext = ".mdl")
#	Blender.Window.FileSelector(file_callback, fname)
	global nCriticalExceptions
	nCriticalExceptions = 0

	global mdl
	mdl = ExportSGL()
	
	registry_load()

	# perform some user input
	result = Blender.Draw.PupMenu("Export for SEGA Saturn SGL:%t|Use default settings|Customize default settings")
	if result == 2:
		result = Blender.Draw.PupIntInput("MaxTexWidth", mdl.maxWidth, 8, 504)
		if result != None:
			mdl.maxWidth = result
		result = Blender.Draw.PupIntInput("MaxTexHeight", mdl.maxHeight, 1, 255)
		if result != None:
			mdl.maxHeight = result
		result = Blender.Draw.PupIntInput("MaxTexMem", mdl.maxTexels*2/1024, 1, 512)
		if result != None:
			mdl.maxTexels = result*1024/2
		result = Blender.Draw.PupMenu("Select texture filter%t|Nearest Neighbor|Bilinear") # if no %xN int is set, indices start from 1
		if result == 1:
			mdl.bilinear = 0
		elif result == 2:
			mdl.bilinear = 1
		registry_save()

	print "Using options:\nmaxWidth %s\nmaxHeight %s\nmaxTexels %s\nbilinear %s\n" % (mdl.maxWidth, mdl.maxHeight, mdl.maxTexels, mdl.bilinear)
	# beware: returns immediately
	Blender.Window.FileSelector(file_callback, "Save .MDL", mdl.filename_mdl)







