#include <SGL.H>
#include "SegaSaturn_modified.mdl"
#include "SegaSaturn_modified.txr"


/* controls (are bad, I know)
 * D-PAD, L, R             - change position
 * X,Z                     - move sideways (well, maybe)
 * hold A + D-PAD, L, R    - rotate
 * B,C                     - debug: change object rotation order and sign
 */


// []
// {}

int perm, sign;

// taken from SGL examples, modified like in SPRITE.TXT
extern TEXTURE *FormTbl ;
void set_texture(PICTURE *pcptr , Uint32 NbPicture)
{
	TEXTURE *txptr;

	for(; NbPicture-- > 0; pcptr++){
		txptr = FormTbl + pcptr->texno;
		slDMACopy((void *)pcptr->pcsrc,
			(void *)(SpriteVRAM + ((txptr->CGadr) << 3)),
			(Uint32)((txptr->Hsize * txptr->Vsize * 4) >> (pcptr->cmode)));
	}
}


// taken/inspired from Saturn SoundPlayer, modification inspired by AKIRA demo
FIXED light[XYZ];
static ANGLE light_ang[XYZ];
static ANGLE light_angAdd[XYZ];

void initLight()
{
    // LIGHT VECTOR IN SCENE
        // initial light angle
    light_ang[X] = DEGtoANG(0.0);
    light_ang[Y] = DEGtoANG(0.0);
    light_ang[Z] = DEGtoANG(0.0);
        // initial light rotation speed
    light_angAdd[X] = DEGtoANG(1.0);
    light_angAdd[Y] = DEGtoANG(2.0);
    light_angAdd[Z] = DEGtoANG(3.0);
}

void computeLight()
{
//    FIXED light_init[XYZ] = { toFIXED(0.57735026), toFIXED(0.57735026), toFIXED(0.57735026) };
    
	// calculate light vector
		// this method ensures unit length
    slPushUnitMatrix();
    {
        // perform rotation
        slRotX(light_ang[X]);            								 
        slRotY(light_ang[Y]);			   
        slRotZ(light_ang[Z]);
        // transform light vector
//        slCalcVector(light_init, light);	// SoundPlayer method, gimbal lock?
		slCalcPoint(toFIXED(0.0), toFIXED(1.0), toFIXED(0.0), light);	// AKIRA demo method
    }
    slPopMatrix();
    // advance light rotation
    // !!! NO OVERFLOW CHECKING !!!
    light_ang[X] += light_angAdd[X];
    light_ang[Y] += light_angAdd[Y];
    light_ang[Z] += light_angAdd[Z];
}

	
// Note: these two functions assume, that no PDATA at all is used, only XPDATA
//       because they access vntbl, which is only present in XPDATA
void slPutObjectX(OBJECT *objptr)
{
	Uint16 d = Smpc_Peripheral[0].data;
//	if(d & PER_DGT_TA)
	slTranslate(objptr->pos[X], objptr->pos[Y], objptr->pos[Z]) ;
// read bottom to top XYZ, XZY
//  0    1    2    3    4    5
// XYZ, XZY, YZX, YXZ, ZXY, ZYX
//	XZY looks fine for 3 colored pyramids
	ANGLE ang[XYZ];
	ang[X] = objptr->ang[X];
	ang[Y] = objptr->ang[Y];
	ang[Z] = objptr->ang[Z];
	if(sign & 1)
		ang[X] = -ang[X];
	if(sign & 2)
		ang[Y] = -ang[Y];
	if(sign & 4)
		ang[Z] = -ang[Z];
	
	if(4 <= perm && perm <= 5) {
		if(perm == 4) {
			slRotY(ang[Y]);
			slRotX(ang[X]);
		} else {
			slRotX(ang[X]);
			slRotY(ang[Y]);
		}
		slRotZ(ang[Z]);
	}

	if(2 <= perm && perm <= 3) {
		if(perm == 2) {
			slRotX(ang[X]);
			slRotZ(ang[Z]);
		} else {
			slRotZ(ang[Z]);
			slRotX(ang[X]);
		}
		slRotY(ang[Y]);
	}

	if(perm <= 1) {
		if(perm == 0) {
			slRotZ(ang[Z]);
			slRotY(ang[Y]);
		} else {
			slRotY(ang[Y]);
			slRotZ(ang[Z]);
		}
		slRotX(ang[X]);
	}

//	if(d & PER_DGT_TC)
	slScale(objptr->scl[X], objptr->scl[Y], objptr->scl[Z]) ;
	
	if(objptr->pat != NULL){
		if(((XPDATA*)objptr->pat)->vntbl == NULL)
			slPutPolygon(objptr->pat) ;
		else slPutPolygonX((XPDATA*)objptr->pat, light);
	}
}


void slPutAllObjectsX(OBJECT* obptr)
{
	slPushMatrix() ;                 /* Preserve parent matrix */
	{
		slPutObjectX(obptr) ;         /* Display object */
		if(obptr->child != NULL){
			slPutAllObjectsX(obptr->child) ;       /* Display if child is present */
		}
	}
	slPopMatrix() ;

	if(obptr->sibling != NULL){
		slPutAllObjectsX(obptr->sibling) ;     /* Display sibling object */
	}
}





FIXED pos[XYZ] = {toFIXED(0.0), toFIXED(0.0), toFIXED(140.0+50.0)};
FIXED pos_diff[XYZ] = {toFIXED(1.8), toFIXED(1.8), toFIXED(2.8)};
ANGLE rot[XYZ] = {DEGtoANG(0.0), DEGtoANG(0.0), DEGtoANG(0.0)};
ANGLE rot_diff[XYZ] = {DEGtoANG(1.0), DEGtoANG(1.0), DEGtoANG(1.0)};
VECTOR slide, slide_start = {toFIXED(1.8), toFIXED(0.0), toFIXED(0.0)};

void modify32(FIXED *f, FIXED *d, Uint16 p)
{
	if(p & PER_DGT_KL) f[0] -= d[0];
	if(p & PER_DGT_KR) f[0] += d[0];

	if(p & PER_DGT_KU) f[1] -= d[1];
	if(p & PER_DGT_KD) f[1] += d[1];

	if(p & PER_DGT_TL) f[2] -= d[2];
	if(p & PER_DGT_TR) f[2] += d[2];
}
void modify16(ANGLE *f, ANGLE *d, Uint16 p)
{
	if(p & PER_DGT_KL) f[0] -= d[0];
	if(p & PER_DGT_KR) f[0] += d[0];

	if(p & PER_DGT_KU) f[1] -= d[1];
	if(p & PER_DGT_KD) f[1] += d[1];

	if(p & PER_DGT_TL) f[2] -= d[2];
	if(p & PER_DGT_TR) f[2] += d[2];
}


extern	Uint16	TotalPolygons;
extern	Uint16	DispPolygons;


void ss_main(void) 
{
	int frame = 0;
	perm = 0;
	sign = 0;

	slInitSystem(TV_320x240, TexTbl, 1);
	set_texture(PicTbl , PicTblLength);
	slPrint("Tunnel Racing", slLocate(14,1));
	slPrint("TotalPolygons", slLocate(25,3));
	slPrint("DispPolygons", slLocate(25,5));
	slPrint("Frame", slLocate(25,7));

	// display gouraud usage
	if (GrStaticTblLength == 0)
		slPrint("No static gouraud shading used", slLocate(2, 28));
	else {
		slDMACopy(GrStaticTbl,GrStaticDmaStart,sizeof(GrStaticTbl));
		
		slPrintFX(GrStaticTblLength << 16, slLocate(31, 28));
		slPrint("Static gouraud table length:", slLocate(2, 28));
	}

	if (GrRealTblLength == 0)
		slPrint("No real time gouraud shading used", slLocate(2, 29));
	else {
		slInitGouraud(GrRealWorkTbl, GrRealTblLength, GrReal(0), GrRealWorkSpc);
		slIntFunction(slGouraudTblCopy);
		
		slPrintFX(GrRealTblLength << 16, slLocate(31, 29));
		slPrint("Real time gouraud table length:", slLocate(2, 29));
	}

	initLight();
	// register flat shading light vector
	slLight(light);
	
	while(1) {
		computeLight();
		// inputs
		Uint16 d = ~Smpc_Peripheral[0].data;

		if(d & PER_DGT_TA)
			modify16(rot, rot_diff, d);
		else
			modify32(pos, pos_diff, d);
		
		// slide to side
		if(d & (PER_DGT_TX | PER_DGT_TZ)) {
			slPushUnitMatrix();
			slRotZ(rot[Z]);
			slRotY(rot[Y]);
			slRotX(rot[X]);
			slCalcVector(slide_start, slide);
			slPopMatrix();
			if(d & PER_DGT_TX) {
				pos[X] -= slide[X];
				pos[Y] -= slide[Y];
				pos[Z] -= slide[Z];
			}
			if(d & PER_DGT_TZ) {
				pos[X] += slide[X];
				pos[Y] += slide[Y];
				pos[Z] += slide[Z];
			}
		}


//		if(~Smpc_Peripheral[0].push & PER_DGT_TX)
//			mode = ((mode+1) & 8) | (mode & ~8);
//		if(~Smpc_Peripheral[0].push & PER_DGT_TY)
//			mode = ((mode+(1 << 4)) & (8 << 4)) | (mode & ~(8 << 4));
		if(~Smpc_Peripheral[0].push & PER_DGT_TC) {
			if(++perm > 5)
				perm = 0;
		}
		if(~Smpc_Peripheral[0].push & PER_DGT_TB) {
			if(++sign > 7)
				sign = 0;
		}

		// draw
	    slUnitMatrix(CURRENT);
		slTranslate(pos[X], pos[Y], pos[Z]);
		slRotZ(rot[Z]);
		slRotY(rot[Y]);
		slRotX(rot[X]);

		
//		if((~d) & PER_DGT_TB)
			slScale(toFIXED(10.0), toFIXED(10.0), toFIXED(10.0));
// yabause crash
//		slPutObject(first_root_object);
// works:
		//slPutPolygon((PDATA *)xpdata_Circle_001);
//		slPutPolygon((PDATA *)xpdata_Cone);
//		slPutObject(object_Cone);
//		slPutObject(object_Cylinder);	// siblings (and their childs) not displayd
//		slPutObject(object_Cube_B0);	// only that one cube is displayed!

//		slPutAllObjectsX(first_root_object);

//		slPutObjectX(object_Cone);	// top up, okay
//		slPutAllObjectsX(object_Cube_B0);
		
//		slPutObjectX(object_Cone3ColObj); // single object with no object transformations is okay
//		slPutAllObjectsX(object_Cone3ColObj);	// three pyramids okay, with XZY

		slPutAllObjectsX(first_root_object);

// to verify static gouraud, it works
//		slPutObject(object_Cone);

		slPrintHex(frame++, slLocate(25+3,8));
		slPrintFX(TotalPolygons<<16, slLocate(30,4));
		slPrintFX(DispPolygons<<16, slLocate(30,6));
		slPrint("light", slLocate(25,10));
		slPrintFX(light[X], slLocate(30,10));
		slPrintFX(light[Y], slLocate(30,11));
		slPrintFX(light[Z], slLocate(30,12));
//#define DEGtoANG(d)		((ANGLE)((65536.0 * (d)) / 360.0))
#define ANGtoDEG(d)		toFIXED(((float)(d)) * 360.0 / 65536.0)
		slPrint("light_ang", slLocate(25,13));
		slPrintFX(ANGtoDEG(light_ang[X]), slLocate(30,13));
		slPrintFX(ANGtoDEG(light_ang[Y]), slLocate(30,14));
		slPrintFX(ANGtoDEG(light_ang[Z]), slLocate(30,15));
		slPrint("perm", slLocate(25,16));
		slPrintHex(perm, slLocate(25+3,16));
		slPrint("sign", slLocate(25,17));
		slPrintHex(sign, slLocate(25+3,17));
		
		slSynch();
	}

}
